/*
          LTR216.
            ,    
           :
    ltr216_recv  slot crate_serial srvip
    :
        slot         -   ( 1  16)
        crate_serial -    (   )
        srvip        - ip-  ltrd  LtrServer (     )
     :    ,       

              .
        - ( - )   
      .

         ,   
       .

           Windows   CTRL+C  Linux

      VisualStudio:
          Visual Studio,     
    ( (Project) ->  (Properties) ->   (Configuration Properties)
    -> /++ ->  (General) ->    (Additional Include Directories))
            (ltr216api.h    
         ltr/include   )
        .lib   <  >/lib/msvc
    ( (Project) ->  (Properties) ->   (Configuration Properties) ->
     (Linker) ->  (General) ->    (Additional Library Directories)).

    !:      Visual Studio    
       ,    :
      (File) ->    (Advanced Save Options)...
        (Encoding)   (UTF8,  )/Unicode (UTF-8 with signature)
        .
*/

#include "ltr/include/ltr216api.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef _WIN32
#include <locale.h>
#include <conio.h>
#else
#include <signal.h>
#include <unistd.h>
#endif


#define LCH_CNT             (3)
/*    ,    */
#define RECV_BLOCK_FRAMES_CNT  (100/LCH_CNT)
/*       */
#define RECV_TOUT           (1000)


typedef struct {
    int slot;
    const char *serial;
    DWORD addr;
} t_open_param;


/*      */
static int f_out = 0;

#ifndef _WIN32
/*     Linux */
static void f_abort_handler(int sig) {
    f_out = 1;
}
#endif


/*    .   ,  
 *   :
 * 1  -   ( 1  16)
 * 2  -   
 * 3  - ip-  */
static int f_get_params(int argc, char** argv, t_open_param* par) {
    int err = 0;
    par->slot = LTR_CC_CHNUM_MODULE1;
    par->serial = "";
    par->addr = LTRD_ADDR_DEFAULT;


    if (argc > 1)
        par->slot = atoi(argv[1]);
    if (argc > 2)
        par->serial = argv[2];
    if (argc > 3) {
        int a[4],i;
        if (sscanf(argv[3], "%d.%d.%d.%d", &a[0], &a[1], &a[2], &a[3])!=4) {
            fprintf(stderr, "  IP-!!\n");
            err = -1;
        }

        for (i=0; (i < 4) && !err; i++) {
            if ((a[i]<0) || (a[i] > 255)) {
                fprintf(stderr, " IP-!!\n");
                err = -1;
            }
        }

        if (!err) {
            par->addr = (a[0] << 24) | (a[1]<<16) | (a[2]<<8) | a[3];
        }
    }
    return err;
}

/*  ,      */
static void f_print_meas_status(DWORD status) {
    switch (status) {
        case LTR216_MEASSTATUS_NOT_USED: printf(" ."); break;
        case LTR216_MEASSTATUS_OK: printf(""); break;
        case LTR216_MEASSTATUS_NOT_INIT: printf("."); break;
        case LTR216_MEASSTATUS_ADC_OVERRANGE: printf(""); break;
        case LTR216_MEASSTATUS_SHORT: printf(""); break;
        case LTR216_MEASSTATUS_OPEN: printf("."); break;
        case LTR216_MEASSTATUS_BAD_VALUE_RANGE: printf(" ."); break;
        case LTR216_MEASSTATUS_CANT_CALC: printf(" ."); break;
        default: printf("%d", status); break;
   }
}

/*   ,   ,      */
static void f_print_meas_valstate(const TLTR216_MEASSTATE *meas) {
    if (meas->ValueValid) {
        printf("%.3f", meas->Value);
    } else {
        f_print_meas_status(meas->Status);
    }
}

/*     Uref (     
 * UrefR)   */
static void f_print_uref(const TLTR216_DATA_STATUS *mstatus) {
    printf("Uref = ");
    f_print_meas_valstate(&mstatus->Uref);
    if (mstatus->UrefR.Status != LTR216_MEASSTATUS_NOT_USED) {
        printf(", UrefR = ");
        f_print_meas_valstate(&mstatus->UrefR);
    }
}


int main(int argc, char** argv) {
    INT err = LTR_OK;
    TLTR216 hltr216;
    t_open_param par;
#ifndef _WIN32
    struct sigaction sa;
    memset(&sa, 0, sizeof(sa));
    /*   Linux      ,
           */
    sa.sa_handler = f_abort_handler;
    sigaction(SIGTERM, &sa, NULL);
    sigaction(SIGINT, &sa, NULL);
    sigaction(SIGABRT, &sa, NULL);
#endif

#ifdef _WIN32
    /*         Windows*/
    setlocale(LC_ALL, "");
#endif

    err = f_get_params(argc, argv, &par);

    if (err==LTR_OK) {
        LTR216_Init(&hltr216);

        /*     */
        err = LTR216_Open(&hltr216, par.addr, LTRD_PORT_DEFAULT, par.serial, par.slot);
        if (err!=LTR_OK) {
            fprintf(stderr, "     .  %d (%s)\n",
                    err, LTR216_GetErrorString(err));
        } else {
            INT close_err;
            BYTE ch_idx;

            /*     */
            printf("  !   : \n");
            printf("       = %s\n", hltr216.ModuleInfo.Name);
            printf("        = %s\n", hltr216.ModuleInfo.Serial);
            printf("           = %d\n", hltr216.ModuleInfo.VerFPGA);
            fflush(stdout);

            /*   */
            /*   */
            hltr216.Cfg.AdcSwMode = LTR216_ADC_SWMODE_MULTICH_SYNC;
            /*   - 10  */
            LTR216_FillSyncFreqDiv(&hltr216, 10000, NULL);
            /*    - 20  (  7-10 ) */
            hltr216.Cfg.AdcMinSwTimeUs = 20;
            /*   ,         */
            LTR216_FillFilterParams(&hltr216, 0, NULL, NULL);

            /*     */
            hltr216.Cfg.LChCnt = LCH_CNT;
            for (ch_idx = 0; ch_idx < hltr216.Cfg.LChCnt; ch_idx++) {
                hltr216.Cfg.LChTbl[ch_idx].PhyChannel = ch_idx;
                hltr216.Cfg.LChTbl[ch_idx].Range = LTR216_RANGE_35;
            }

            /*    -   */
            hltr216.Cfg.BgMeas = LTR216_BG_MEASGROUP_ALL;

            /*    - 25  */
            LTR216_FillISrcCode(&hltr216, 25, NULL);
            /* 4-      */
            hltr216.Cfg.Ch16ForUref = FALSE;
            /*   */
            hltr216.Cfg.TareEnabled = FALSE;

            
            

            /*     */
            err = LTR216_SetADC(&hltr216);
            if (err != LTR_OK) {
                fprintf(stderr, "    .  %d (%s)\n",
                        err, LTR216_GetErrorString(err));
            } else {
                printf("   .\n   = %.2f \n    %.2f\n  : %d\n",
                       hltr216.State.AdcFreq, hltr216.State.FrameFreq, hltr216.State.FrameWordsCount);
                fflush(stdout);
            }

            /*     */
            if (err == LTR_OK) {
                TLTR216_DATA_STATUS status;
                LTR216_DataStatusInit(&status);

                err = LTR216_InitMeasParams(&hltr216, LTR216_INIT_MEAS_ALL,
                                            LTR216_BG_MEASGROUP_UREF_CHECK, &status);
                if (err != LTR_OK) {
                    fprintf(stderr, "     .  %d (%s)\n",
                            err, LTR216_GetErrorString(err));
                } else {
                    printf("    \n");
                }

                /*      , 
                 *   Uref */
                f_print_uref(&status);
                printf("\n");
                fflush(stdout);

                if (err==LTR_OK) {
                    DWORD recvd_blocks=0;
                    /*   ,      */
                    INT recv_data_cnt = RECV_BLOCK_FRAMES_CNT*hltr216.Cfg.LChCnt;
                    /*  ,    ,
                     *     */
                    INT   recv_wrd_cnt = RECV_BLOCK_FRAMES_CNT*hltr216.State.FrameWordsCount;
                    /*    */
                    DWORD  *rbuf = (DWORD*)malloc(recv_wrd_cnt*sizeof(rbuf[0]));
                    double *data = (double *)malloc(recv_data_cnt*sizeof(data[0]));
                    TLTR216_DATA_CH_STATUS *chStatusList = (TLTR216_DATA_CH_STATUS*)
                            malloc(hltr216.Cfg.LChCnt * sizeof(chStatusList[0]));


                    if ((rbuf==NULL) || (data==NULL) || (chStatusList==NULL)) {
                        fprintf(stderr, "  !\n");
                        err = LTR_ERROR_MEMORY_ALLOC;
                    }

                    if (err==LTR_OK) {
                        LTR216_DataChannelsStatusInit(chStatusList, hltr216.Cfg.LChCnt);

                        /*    */
                        err=LTR216_Start(&hltr216);
                        if (err!=LTR_OK) {
                            fprintf(stderr, "    !  %d (%s)\n",
                                    err, LTR216_GetErrorString(err));
                        }
                    }

                    if (err==LTR_OK) {
                        printf("  .    %s\n",
       #ifdef _WIN32
                               " "
       #else
                               "CTRL+C"
       #endif
                               );
                        fflush(stdout);
                    }

                    /*        
                     *    */
                    while (!f_out && (err==LTR_OK)) {
                        INT recvd;
                        /*         */
                        DWORD tout = RECV_TOUT + (DWORD)(1000.*RECV_BLOCK_FRAMES_CNT/
                                                         hltr216.State.FrameFreq + 1);
                        /*    .  */
                        recvd = LTR216_Recv(&hltr216, rbuf, NULL, recv_wrd_cnt, tout);

                        /*       */
                        if (recvd < 0) {
                            err = recvd;
                            fprintf(stderr, "  .  %d:%s\n",
                                    err, LTR216_GetErrorString(err));
                        } else if (recvd != recv_wrd_cnt) {
                            fprintf(stderr, "  .  %d,  %d\n",
                                    recv_wrd_cnt, recvd);
                            err = LTR_ERROR_RECV_INSUFFICIENT_DATA;
                        } else {
                            err = LTR216_ProcessData(&hltr216, rbuf, data, &recvd,
                                                     LTR216_PROC_FLAG_MEAS_UNBALANCE,
                                                     &status, chStatusList);
                            if (err != LTR_OK) {
                                fprintf(stderr, "  .  %d:%s\n",
                                        err, LTR216_GetErrorString(err));
                            } else {
                                DWORD ch;

                                recvd_blocks++;
                                /*    */
                                printf(" %d ( = %d, ", recvd_blocks, recvd);
                                /*   uref   */
                                f_print_uref(&status);
                                printf("):");

                                /*       
                                 *   */
                                for (ch=0; (ch < hltr216.Cfg.LChCnt) ; ch++) {
                                    /*   
                                     * ,      
                                     * .    
                                     *      
                                     *       
                                     *    */
                                    if (chStatusList[ch].Ures.ValueValid) {
                                        printf("%8.2f", 1000.*data[ch]);
                                    } else {
                                        f_print_meas_status(chStatusList[ch].Ures.Status);
                                    }

                                    if (ch==(hltr216.Cfg.LChCnt-1)) {
                                        printf("\n");
                                    } else {
                                        printf(",  ");
                                    }
                                }
                                fflush(stdout);
                            }
                        }

#ifdef _WIN32
                        /*      */
                        if (err==LTR_OK) {
                            if (_kbhit())
                                f_out = 1;
                        }
#endif
                    } //while (!f_out && (err==LTR_OK))

                    /*    ,    */
                    if (hltr216.State.Run) {
                        INT stop_err = LTR216_Stop(&hltr216);
                        if (stop_err!=LTR_OK) {
                            fprintf(stderr, "    .  %d:%s\n",
                                    stop_err, LTR216_GetErrorString(stop_err));
                            if (err==LTR_OK)
                                err = stop_err;
                        } else {
                            printf("  .\n");
                        }
                    }

                    free(rbuf);
                    free(data);
                    free(chStatusList);

                }
            }


            /*     */
            close_err = LTR216_Close(&hltr216);
            if (close_err!=LTR_OK) {
                fprintf(stderr, "     .  %d:%s\n",
                        close_err, LTR216_GetErrorString(close_err));
                if (err==LTR_OK)
                    err = close_err;
            } else {
                printf("    .\n");
            }
        }
    }

    return err;
}
